/******************************************************************************
 *
 * dumpstream.c - Dump NTFS files attributes. Part of the DirTools project.
 *
 * $Revision: 1.3 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <inttypes.h>
#include <unistd.h>
#include <getopt.h>
#ifdef _WIN32
#include <windows.h>
#else /* _WIN32 */
#include "wincompat.h"
#endif /* _WIN32 */

#include "dt_error.h"
#include "structure.h"
#include "diskio.h"
#include "output.h"
#include "utility.h"
#include "macros.h"
#include "dt_locale.h"

#ifndef _fcloseall
int _fcloseall (void);
#endif


/**********************************************************************
 *
 *
 *
 *
 *
 *
 *********************************************************************/
void
Usage (const char *pProgName)
{
#ifdef _WIN32
     fprintf (stderr,
              "Usage: %s [-H][-h][-t attribute_type][-n attribute_name |"
              "-c attribute_number][-D][-d debug_file][-x error_file]"
              "[-o output_file][-8] drive_letter MFT_index\n",
              pProgName);
#else /* _WIN32 */
     fprintf (stderr,
              "Usage: %s [-H][-h][-t attribute_type][-n attribute_name |"
              "-c attribute_number][-D][-d [debug_file]][-x error_file]"
              "[-o output_file][-s skip][-8] file MFT_index\n",
              pProgName);
#endif /* _WIN32 */
}

/**********************************************************************
 *
 *
 *
 *
 *
 *
 *********************************************************************/
void
DisplayHelp (const char *pProgName)
{
     fprintf (stderr, "%s: Dump NTFS files attributes\n\n", pProgName);
     Usage (pProgName);
     fprintf (stderr, "\nPackage version: %s\n", PACKAGE_STRING);
     fprintf (stderr, "\n-h : help\n");
     fprintf (stderr, "-H : hex instead of raw dump\n");
     fprintf (stderr, "-t : specify the attribute type\n");
     fprintf (stderr, "     default output is data attribute stream\n");
     fprintf (stderr, "-n : specify a name for a given type of attribute\n");
     fprintf (stderr, "     usefull to get alternate data streams\n");
     fprintf (stderr,
              "-c : specify the occurence of the attribute type to output\n");
     fprintf (stderr, "     this option is not compatible with -n\n");
     fprintf (stderr, "-d, -D : debug mode, -D dump each MFT record read\n");
#ifndef _WIN32
     fprintf (stderr, "         default output is sdterr\n");
#endif
     fprintf (stderr, "-x : output error to file instead of stderr\n");
     fprintf (stderr, "-o : results sent to file instead of stdout\n");
#ifndef _WIN32
     fprintf (stderr, "-s : skip bytes when reading file\n");
#endif /* _WIN32 */
     fprintf (stderr, "-8 : output non-ascii chars as UTF-8\n");
     fprintf (stderr, "\nType values :\n");
     fprintf (stderr, " 0x10  : standard\n");
     fprintf (stderr, " 0x20  : attribute list\n");
     fprintf (stderr, " 0x30  : filename\n");
     fprintf (stderr, " 0x40  : volume version(NT), object id (2K)\n");
     fprintf (stderr, " 0x50  : security\n");
     fprintf (stderr, " 0x60  : volume name\n");
     fprintf (stderr, " 0x70  : volume information\n");
     fprintf (stderr, " 0x80  : data\n");
     fprintf (stderr, " 0x90  : index root\n");
     fprintf (stderr, " 0xA0  : index allocation\n");
     fprintf (stderr, " 0xB0  : bitmap\n");
     fprintf (stderr, " 0xC0  : symbolic link(NT), reparse point (2K)\n");
     fprintf (stderr, " 0xD0  : HPFS extended information\n");
     fprintf (stderr, " 0xE0  : HPFS extended\n");
     fprintf (stderr, " 0xF0  : property set\n");
     fprintf (stderr, " 0x100 : logged utility stream (2K)\n");

}

/**********************************************************************
 *
 *
 *
 *
 *
 *
 *********************************************************************/
int
main (int argc, char **argv)
{
     int iArg, iOptInd, iErrorLevel = OK_ERROR, iCurrentError;
     NTFS_VOL_INFO nvVol;
     PNTFS_MFT_ENTRY pEntry;
     PNTFS_ATTRIBUTE pAttr = NULL;
     wchar_t wcName[ATTRIBUTE_MAX_NAME_LENGTH + 1];
     wchar_t *pName = NULL;
     uint64_t ulNum;
     int bHexDump;
     FILE *fOutput;
     uint32_t dwType = 0;
     uint16_t wNum = 0, wFound = 0, wCount;
     char *pOutputFilename = NULL, *pDebugFilename = NULL, *pErrorFilename =
          NULL;
     DT_LOCALE_TYPE mylocale = DT_SET_LOCALE;
     static struct option long_opts[] = {
          {"help", 0, 0, 'h'},
          {"Hex", 0, 0, 'H'},
          {"Dump", 0, 0, 'D'},
	  {"utf8", 0, 0, '8'},
#ifndef _WIN32
          {"skip", 1, 0, 's'},
#endif
          {"type", 1, 0, 't'},
          {"name", 1, 0, 'n'},
          {"count", 1, 0, 'c'},
#ifdef _WIN32
          {"debug", 1, 0, 'd'},
#else
          {"debug", 2, 0, 'd'},
#endif
          {"error-output", 1, 0, 'x'},
          {"output", 1, 0, 'o'},
          {0, 0, 0, 0}
     };
#ifdef _WIN32
#define OPTSTRING "hHD8t:n:c:d:x:o:"
#else
#define OPTSTRING "hHD8s:t:n:c:d::x:o:"
#endif

     FErrorOutput = stderr;
     FIoDebug = NULL;
     BIoDump = FALSE;
     fOutput = stdout;
     bHexDump = FALSE;
     ulNum = 0;

     wcName[0] = L'\0';
#ifndef _WIN32
     nvVol.ulSkip = 0;
#endif /* _WIN32 */

     dt_setlocale("");

     while ((iArg =
             getopt_long (argc, argv, OPTSTRING, long_opts,
                          &iOptInd)) != -1) {
          switch (iArg) {
          case 'h':
               DisplayHelp (argv[0]);
               _fcloseall ();
               return OK_ERROR;
               break;

          case 'H':
               bHexDump = TRUE;
               break;

	  case '8':
	       mylocale = DT_SET_UTF8;
	       break;

          case 't':
               if (optarg)
/*              dwType = _ttol(optarg); */
                    dwType = (uint32_t) integer_arg (optarg);
               else {
                    Alarm (TA_PROG_FATAL,
                           "%s : requested attribute type not found",
                           argv[0]);
                    _fcloseall ();
                    return FATAL_ERROR;
               }
               break;

          case 'n':
               if (optarg) {
		    dt_mbstowcs(wcName, optarg, ATTRIBUTE_MAX_NAME_LENGTH);
                    wcName[ATTRIBUTE_MAX_NAME_LENGTH] = L'\0';
               } else {
                    Alarm (TA_PROG_FATAL,
                           "%s : requested attribute name not found",
                           argv[0]);
                    _fcloseall ();
                    return FATAL_ERROR;
               }
               break;

          case 'c':
               if (optarg)
                    wNum = atoi (optarg);
               else {
                    Alarm (TA_PROG_FATAL,
                           "%s : requested attribute occurence not found",
                           argv[0]);
                    _fcloseall ();
                    return FATAL_ERROR;
               }
               break;

#ifndef _WIN32
          case 's':
               if (optarg)
                    nvVol.ulSkip = integer_arg (optarg);
               else {
                    Alarm (TA_PROG_FATAL,
                           "%s : skip value not found", argv[0]);
                    _fcloseall ();
                    return FATAL_ERROR;
               }
               break;
#endif /* _WIN32 */

          case 'o':
               MANDATORYFILENAME (optarg, pOutputFilename, "output");
               break;

          case 'x':
               MANDATORYFILENAME (optarg, pErrorFilename, "error");
               break;

          case 'D':
               BIoDump = TRUE;
               break;

          case 'd':
               if (optarg)
                    pDebugFilename = optarg;
               else
                    FIoDebug = stderr;
               break;

          default:
               Usage (argv[0]);
               _fcloseall ();
               return FATAL_ERROR;
               break;
          }                     /* switch(iArg) */
     }                          /* while((iArg = getopt_long(argc, argv, ... */

     if (wcName[0] && wNum) {
          Alarm (TA_PROG_FATAL, "%s : -n and -s are not compatible", argv[0]);
          _fcloseall ();
          return FATAL_ERROR;
     }

     if (optind != (argc - 2)) {        /* 2 non-option arguments */
          Usage (argv[0]);
          Alarm (TA_PROG_FATAL, "%s : wrong number of arguments", argv[0]);
          _fcloseall ();
          return FATAL_ERROR;
     }
#ifdef _WIN32
     if (argv[optind][0] >= TEXT ('a') && argv[optind][0] <= TEXT ('z')) {
          nvVol.tDrive = argv[optind][0] - TEXT ('a') + TEXT ('A');
     } else if (argv[optind][0] >= TEXT ('A') &&
                argv[optind][0] <= TEXT ('Z')) {
          nvVol.tDrive = argv[optind][0];
     } else {
          Alarm (TA_PROG_FATAL,
                 "%s : not a valid drive %c", argv[0], argv[optind][0]);
          _fcloseall ();
          return FATAL_ERROR;
     }
#else /* _WIN32 */
     nvVol.file = argv[optind];
#endif /* _WIN32 */

     /* Get the MFT index value */
     optind++;
     if (argv[optind])
/*    ulNum = _ttoi64(argv[optind]); */
          ulNum = integer_arg (argv[optind]);
     else {
          Alarm (TA_PROG_FATAL, "%s : MFT index value not found", argv[0]);
          _fcloseall ();
          return FATAL_ERROR;
     }

     set_wc_output_function(mylocale);

     OUTPUTFILE (pOutputFilename, fOutput, "output");

     OUTPUTFILE (pDebugFilename, FIoDebug, "debug");
     if (BIoDump && FIoDebug == NULL)
          FIoDebug = stderr;

     OUTPUTFILE (pErrorFilename, FErrorOutput, "error");

     iErrorLevel = OpenVolume (&nvVol);

     if (iErrorLevel == OK_ERROR) {
          if (!dwType)
               dwType = nvVol.dwAtData; /* Init dwType if not specified */

/*      if (wcName[0] || (dwType == nvVol.dwAtData)) */
          if (wcName[0])
               pName = wcName;

          iErrorLevel = InitEntry (&nvVol, &pEntry, ulNum, FALSE);
          if (iErrorLevel == OK_ERROR) {
               if (wNum) {
                    for (wCount = 0; wCount < pEntry->wAttrCount; wCount++) {
                         if (pEntry->pAttr[wCount].dwType == dwType)
                              wFound++;

                         if (wFound == wNum) {
                              pAttr = &pEntry->pAttr[wCount];
                              break;
                         }
                    }
               } else           /* if (wNum) */
                    pAttr = FindAttribute (pEntry, dwType, pName);

               if (pAttr) {
                    if (bHexDump)
                         ProcessStream (pEntry, pAttr, &ProcessHexBytes,
                                        (void *) fOutput);
                    else {
#ifdef _WIN32
                         /* to avoid CRLF conversion */
                         _setmode (_fileno (fOutput), _O_BINARY);
#endif /* _WIN32 */
                         ProcessStream (pEntry, pAttr, &ProcessRawBytes,
                                        (void *) fOutput);
                    }
               } else {
                    Alarm (TA_PROG_ERR,
                           "%s : attribute 0x%" PRIx32
                           " from base record %" PRIu64 " not found",
                           argv[0], dwType, ulNum);
                    return SIMPLE_ERROR;
               }

          }
          /*  if (iErrorLevel == OK_ERROR) */
     }
     iCurrentError = FreeEntry (pEntry);

     iErrorLevel = max (iCurrentError, iErrorLevel);

     iCurrentError = CloseVolume (&nvVol);

     iErrorLevel = max (iErrorLevel, iCurrentError);

     _fcloseall ();
     return iErrorLevel;
}
