/******************************************************************************
 *
 * ntdd.c - Transmit NT file over TCP/IP. Part of the DirTools project.
 *
 * $Revision: 1.1.1.1 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <getopt.h>
#ifdef _WIN32
#include <windows.h>
#include <winsock.h>
#else
#define closesocket close
#define O_BINARY 0
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#define DEFAULT_FILE "\\\\.\\C:"
#define DEFAULT_SIZE 1048576
#define DEFAULT_PORT 1313

extern int errno;

void
print_usage (const char *prog)
{
     fprintf (stderr, "Usage: %s [--help][--buffer buffer_size]"
              "[--port remote_port][--file file_to_copy]"
              " remote_host\n", prog);
}

void
print_help (const char *prog)
{
     fprintf (stderr, "%s: Send an NT/2k/XP file over TCP/IP\n", prog);
     print_usage (prog);
     fprintf (stderr, "\t--help, -h\tthis help\n");
     fprintf (stderr, "\t--port, -p\tremote port (default 1313)\n");
     fprintf (stderr, "\t--buffer, -b\tsize of the read buffer ;"
              " must be a multiple of the disk sector size,"
              " usually 512 (defaut 1048576)\n");
     fprintf (stderr, "\t--file, -f\tfile to read: \\\\.\\D:,"
              " \\\\.\\E:, \\\\.\\PHYSICALDRIVE0, \\\\.\\TAPE0,..."
              " (defaut \\\\.\\C:)\n");
}

int
main (int argc, char *argv[])
{
     int fd, sd, s_sent;
     size_t size = DEFAULT_SIZE, s_read;
     unsigned char *buf = NULL;
     uint64_t total = 0;
     struct hostent *ptrh;
     struct protoent *ptrp;
     struct sockaddr_in sad;
     int port = DEFAULT_PORT, opt, opt_index;
     char *host;
     char *disque = DEFAULT_FILE;
     static struct option long_opts[] = {
          {"help", 0, 0, 'h'},
          {"buffer", 1, 0, 'b'},
          {"port", 1, 0, 'p'},
          {"file", 1, 0, 'f'},
          {0, 0, 0, 0}
     };
#ifdef _WIN32
     WSADATA wsaData;

     WSAStartup (0x101, &wsaData);
#endif

     memset ((char *) &sad, 0, sizeof (sad));
     sad.sin_family = AF_INET;

     while ((opt =
             getopt_long (argc, argv, "hb:p:f:", long_opts,
                          &opt_index)) > 0) {
          switch (opt) {
          case 'h':
               print_help (argv[0]);
               exit (EXIT_SUCCESS);
               break;
          case 'p':
               port = atoi (optarg);
               break;
          case 'b':
               size = atoi (optarg);
               if (size % 512)
                    fprintf (stderr, "Warning %u is not multiple of 512 !\n",
                             size);
               break;
          case 'f':
               disque = optarg;
               break;
          default:
               print_usage (argv[0]);
               exit (EXIT_FAILURE);
               break;
          }
     }

     if (!argv[optind]) {
          print_usage (argv[0]);
          exit (EXIT_FAILURE);
     }
     host = argv[optind];

     /* check port value and assign to socket */
     if (port > 0) {
          sad.sin_port = htons ((u_short) port);
     } else {
          fprintf (stderr, "Bad port: %u\n", port);
          exit (EXIT_FAILURE);
     }
     if (size <= 0) {
          fprintf (stderr, "Unknown buffer size: %u\n", size);
          exit (EXIT_FAILURE);
     }
     /* convert host name to IP address */
     ptrh = gethostbyname (host);
     if (((char *) ptrh) == NULL) {
          fprintf (stderr, "Unknown host: %s\n", host);
          exit (EXIT_FAILURE);
     }
     memcpy (&sad.sin_addr, ptrh->h_addr, ptrh->h_length);
     /* map TCP to protocol number */
     if (((int) (ptrp = getprotobyname ("tcp"))) == 0) {
          fprintf (stderr, "Can't find tcp protocol number\n");
          exit (EXIT_FAILURE);
     }
     /* read buffer allocation */
     buf = malloc (size);
     if (!buf) {
          fprintf (stderr, "Memory failure\n");
          exit (EXIT_FAILURE);
     }

     /* volume opening */
     fd = open (disque, O_RDONLY | O_BINARY);
     if (fd == -1) {
          fprintf (stderr, "Can't open %s: %s (%u)\n", disque,
                   strerror (errno), errno);
          free (buf);
          exit (EXIT_FAILURE);
     }

     /* setmode(fileno(stdout), O_BINARY); */

     /* create socket */
     if ((sd = socket (PF_INET, SOCK_STREAM, ptrp->p_proto)) < 0) {
          fprintf (stderr, "Can't create socket: %s (%u)\n", strerror (errno),
                   errno);
          free (buf);
          close (fd);
          exit (EXIT_FAILURE);
     }
     if (connect (sd, (struct sockaddr *) &sad, sizeof (sad)) < 0) {
          fprintf (stderr, "Can't connect");
          if (errno)
               fprintf (stderr, ": %s (%u)\n", strerror (errno), errno);
          else
               fprintf (stderr, "\n");
          free (buf);
          closesocket (sd);
          close (fd);
          exit (EXIT_FAILURE);
     }

     /* boucle principale */
     while ((s_read = read (fd, buf, size)) > 0) {
          total += s_read;
          s_sent = send (sd, buf, s_read, 0);
          if (s_sent < s_read) {
               if (s_sent == -1) {
                    fprintf (stderr, "Transport failure: %s (%u)",
                             strerror (errno), errno);
                    closesocket (sd);
                    close (fd);
                    free (buf);
                    exit (EXIT_FAILURE);
               } else {
                    fprintf (stderr, "Buffer partly transmitted"
                             ": %x < %x\n", s_sent, s_read);
               }
          }
          if (!(total & 0xFFF))
               printf ("Bytes sent : %" PRIu64 "\r", total);
     }
     if (s_read == -1)
          fprintf (stderr, "No more bytes to read %s (%u)\n",
                   strerror (errno), errno);
     printf ("Total read/sent : %" PRIu64 "              \n", total);

     free (buf);
     closesocket (sd);
     close (fd);
     return EXIT_SUCCESS;
}
